@@ -13,7 +13,7 @@ from .block import CodeBlockAnalyser, SourceTransform, link_html
 from .directive import RemoveExtensionVisitor
 from .cache import DataCache
 from .resolve import resolve_location, CouldNotResolve
-from ..parse import Name, NameBreak
+from ..parse import Name
 from ..warn import logger, warn_type
 
 
@@ -132,9 +132,11 @@ class SphinxCodeAutoLink:
         doctree.walkabout(visitor)
         self.cache.transforms[visitor.current_document] = visitor.source_transforms
 
-    @staticmethod
-    def merge_environments(app, env, docnames, other):
+    def merge_environments(self, app, env, docnames, other):
         """Merge transform information."""
+        if self.do_nothing:
+            return
+
         env.sphinx_codeautolink_transforms.update(
             other.sphinx_codeautolink_transforms
         )
@@ -192,14 +194,19 @@ class SphinxCodeAutoLink:
         for transform in transforms:
             filtered = []
             for name in transform.names:
-                if name.import_components[-1] == NameBreak.call:
-                    continue  # empty call target (2 calls in a row)
+                if not name.code_str:
+                    continue  # empty transform target (2 calls in a row)
                 try:
                     key = resolve_location(name, self.inventory)
                 except CouldNotResolve:
                     if self.warn_failed_resolve:
+                        path = '.'.join(name.import_components).replace('.()', '()')
+                        msg = (
+                            f'Could not resolve {self._resolve_msg(name)}'
+                            f' using path `{path}`'
+                        )
                         logger.warning(
-                            f'Could not resolve {self._resolve_msg(name)}',
+                            msg,
                             type=warn_type,
                             subtype='failed_resolve',
                             location=(doc, transform.doc_lineno),
