@@ -12,34 +12,21 @@ from dataclasses import dataclass, field
 
 from .warn import logger, warn_type
 
+HAS_WALRUS = (sys.version_info >= (3, 8))
+HAS_MATCH = (sys.version_info >= (3, 10))
+
 
 def parse_names(source: str, doctree_node) -> List['Name']:
     """Parse names from source."""
     tree = ast.parse(source)
     visitor = ImportTrackerVisitor(doctree_node)
     visitor.visit(tree)
-    return sum([split_access(a) for a in visitor.accessed], [])
+    return visitor.accessed
 
 
-@dataclass
-class PendingAccess:
-    """Pending name access."""
-
-    components: List[ast.AST]
-
-
-@dataclass
-class PendingAssign:
-    """
-    Pending assign target.
-
-    `targets` represent the assignment targets.
-    If a single PendingAccess is found, it should be used to store the value
-    on the right hand side of the assignment. If multiple values are found,
-    they should overwrite any names in the current scope and not assign values.
-    """
-
-    targets: Union[Optional[PendingAccess], List[Optional[PendingAccess]]]
+def linenos(node: ast.AST) -> Tuple[int, int]:
+    """Return lineno and end_lineno safely."""
+    return node.lineno, getattr(node, 'end_lineno', node.lineno)
 
 
 @dataclass
@@ -61,12 +48,51 @@ class Component:
         elif isinstance(node, ast.Attribute):
             name = node.attr
             context = node.ctx.__class__.__name__.lower()
+        elif isinstance(node, ast.arg):
+            name = node.arg
         elif isinstance(node, ast.Call):
             name = NameBreak.call
+        elif HAS_MATCH and isinstance(node, ast.MatchAs):
+            name = node.name
+            context = 'store'
         else:
             raise ValueError(f'Invalid AST for component: {node.__class__.__name__}')
-        end_lineno = getattr(node, 'end_lineno', node.lineno)
-        return cls(name, node.lineno, end_lineno, context)
+        return cls(name, *linenos(node), context)
+
+
+@dataclass
+class PendingAccess:
+    """Pending name access."""
+
+    components: List[Component]
+
+
+@dataclass
+class AssignTarget:
+    """
+    Assign target.
+
+    `elements` represent the flattened assignment target elements.
+    If a single PendingAccess is found, it should be used to store the value
+    on the right hand side of the assignment. If multiple values are found,
+    the assignment contained unpacking, and only overwriting of names should occur.
+    """
+
+    elements: List[Optional[PendingAccess]]
+
+
+@dataclass
+class Assignment:
+    """
+    Representation of an assignment statement.
+
+    - ordinarily one value to a single target
+    - multiple targets when chain assigning (a = b = c)
+    - nested assignments in walruses (a = b := c)
+    """
+
+    targets: List[AssignTarget]
+    value: Union[PendingAccess, 'Assignment', None]
 
 
 class NameBreak(str, Enum):
@@ -92,8 +118,8 @@ class Name:
     code_str: str
     lineno: int
     end_lineno: int
-    context: LinkContext = None
-    resolved_location: str = None
+    context: Optional[LinkContext] = None
+    resolved_location: Optional[str] = None
 
 
 @dataclass
@@ -132,8 +158,10 @@ class Access:
     @property
     def code_str(self):
         """Code representation of components."""
-        breaks = set(NameBreak)
-        return '.'.join(c.name for c in self.components if c.name not in breaks)
+        break_on = set(NameBreak)
+        breaks = [i for i, c in enumerate(self.components) if c.name in break_on]
+        start_ix = breaks[-1] + 1 if breaks else 0
+        return '.'.join(c.name for c in self.components[start_ix:])
 
     @property
     def lineno_span(self) -> Tuple[int, int]:
@@ -142,45 +170,44 @@ class Access:
         max_ = max(c.end_lineno for c in self.components)
         return min_, max_
 
-
-def split_access(access: Access) -> List[Name]:
-    """Split access into multiple names."""
-    split = [access]
-    while True:
-        current = split[-1]
-        for i, comp in enumerate(current.components):
-            if i and comp.name == NameBreak.call:
-                hidden = current.hidden_components + current.components[:i]
-                next_ = Access(
-                    LinkContext.after_call,
-                    current.prior_components,
-                    current.components[i:],
-                    hidden_components=hidden,
-                )
-                current.components = current.components[:i]
-                split.append(next_)
-                break
-        else:
-            break
-    if split[-1].components[-1].name == NameBreak.call:
-        split.pop()
-    return [
-        Name(
-            [c.name for c in s.full_components],
-            s.code_str,
-            *s.lineno_span,
-            context=s.context,
+    @staticmethod
+    def to_name(instance: 'Access') -> Name:
+        """Convert access to name."""
+        return Name(
+            [c.name for c in instance.full_components],
+            instance.code_str,
+            *instance.lineno_span,
+            context=instance.context,
         )
-        for s in split
-    ]
 
-
-@dataclass
-class Assignment:
-    """Assignment of value to name."""
-
-    to: List[PendingAssign]
-    value: Optional[PendingAccess]
+    def split(self) -> List[Name]:
+        """Split access into multiple names."""
+        # Copy to avoid modifying the instance in place
+        items = [Access(
+            context=self.context,
+            prior_components=self.prior_components[:],
+            components=self.components[:],
+            hidden_components=self.hidden_components[:],
+        )]
+        while True:
+            current = items[-1]
+            for i, comp in enumerate(current.components):
+                if i and comp.name == NameBreak.call:
+                    hidden = current.hidden_components + current.components[:i]
+                    next_ = Access(
+                        LinkContext.after_call,
+                        current.prior_components,
+                        current.components[i:],
+                        hidden_components=hidden,
+                    )
+                    current.components = current.components[:i]
+                    items.append(next_)
+                    break
+            else:
+                break
+        if items[-1].components[-1].name == NameBreak.call:
+            items.pop()
+        return [self.to_name(i) for i in items]
 
 
 def track_parents(func):
@@ -192,14 +219,11 @@ def track_parents(func):
     @wraps(func)
     def wrapper(self: 'ImportTrackerVisitor', *args, **kwargs):
         self._parents += 1
-        r: Union[PendingAccess, Assignment, None] = func(self, *args, **kwargs)
+        result = func(self, *args, **kwargs)
         self._parents -= 1
         if not self._parents:
-            if isinstance(r, Assignment):
-                self._resolve_assignment(r)
-            elif isinstance(r, PendingAccess):
-                self._access(r)
-        return r
+            self.dispatch_result(result)
+        return result
     return wrapper
 
 
@@ -213,9 +237,10 @@ class ImportTrackerVisitor(ast.NodeVisitor):
 
     def __init__(self, doctree_node):
         super().__init__()
-        self.accessed: List[Access] = []
+        self.accessed: List[Name] = []
         self.in_augassign = False
         self._parents = 0
+        self._no_split = False
         self.doctree_node = doctree_node
 
         # Stack for dealing with class body pseudo scopes
@@ -228,6 +253,18 @@ class ImportTrackerVisitor(ast.NodeVisitor):
         # Holds references to the values of previous nesting levels.
         self.outer_scopes_stack: List[Dict[str, List[Component]]] = []
 
+    def save_access(self, access: Access) -> None:
+        """Convert Access to Names to store in the visitor for aggregation."""
+        names = access.split() if not self._no_split else [Access.to_name(access)]
+        self.accessed.extend(names)
+
+    @contextmanager
+    def no_split(self):
+        """Disable splitting Accesses."""
+        self._no_split, old = (True, self._no_split)
+        yield
+        self._no_split = old
+
     @contextmanager
     def reset_parents(self):
         """Reset parents state for the duration of the context."""
@@ -235,25 +272,26 @@ class ImportTrackerVisitor(ast.NodeVisitor):
         yield
         self._parents = old
 
+    # Nodes that are excempt from resetting parents in default visit
     track_nodes = (
         ast.Name,
         ast.Attribute,
         ast.Call,
-        ast.Assign,
-        ast.AnnAssign,
     )
-    if sys.version_info >= (3, 8):
+    if HAS_WALRUS:
         track_nodes += (ast.NamedExpr,)
+    if HAS_MATCH:
+        track_nodes += (ast.MatchAs,)
 
     def visit(self, node: ast.AST):
         """Override default visit to track name access and assignments."""
-        if not isinstance(node, self.track_nodes):
-            with self.reset_parents():
-                return super().visit(node)
+        if isinstance(node, self.track_nodes):
+            return super().visit(node)
 
-        return super().visit(node)
+        with self.reset_parents():
+            return super().visit(node)
 
-    def _overwrite(self, name: str):
+    def overwrite_name(self, name: str):
         """Overwrite name in current scope."""
         # Technically dotted values could now be bricked,
         # but we can't prevent the earlier values in the chain from being used.
@@ -263,66 +301,84 @@ class ImportTrackerVisitor(ast.NodeVisitor):
         # as the dotted components of imports are discarded on creating the import.
         self.pseudo_scopes_stack[-1].pop(name, None)
 
-    def _assign(self, local_name: str, components: List[Component]):
-        """Import or assign a name."""
-        self._overwrite(local_name)  # Technically unnecessary unless we follow dots
-        self.pseudo_scopes_stack[-1][local_name] = components
-
-    def _access(self, access: PendingAccess) -> Optional[Access]:
-        components = [Component.from_ast(n) for n in access.components]
-        prior = self.pseudo_scopes_stack[-1].get(components[0].name, None)
+    def assign_name(self, name: str, components: List[Component]):
+        """Import or assign a name to current scope."""
+        # Overwriting technically unnecessary until it properly follows dots
+        self.overwrite_name(name)
+        self.pseudo_scopes_stack[-1][name] = components
 
+    def create_access(
+        self, scope_key: str, new_components: List[Component]
+    ) -> Optional[Access]:
+        """Create access from scope."""
+        prior = self.pseudo_scopes_stack[-1].get(scope_key, None)
         if prior is None:
             return
 
+        access = Access(LinkContext.none, prior, new_components)
+        self.save_access(access)
+        return access
+
+    def resolve_pending_access(self, pending: PendingAccess) -> Optional[Access]:
+        """Resolve and save pending access."""
+        components = pending.components
+
         context = components[0].context
         if context == 'store' and not self.in_augassign:
-            self._overwrite(components[0].name)
+            self.overwrite_name(components[0].name)
             return
 
-        access = Access(LinkContext.none, prior, components)
-        self.accessed.append(access)
+        access = self.create_access(components[0].name, components)
         if context == 'del':
-            self._overwrite(components[0].name)
+            self.overwrite_name(components[0].name)
         return access
 
-    def _resolve_assignment(self, assignment: Assignment):
-        value = assignment.value
-        access = self._access(value) if value is not None else None
+    def resolve_assignment(self, assignment: Assignment) -> Optional[Access]:
+        """Resolve access for assignment values and targets."""
+        access = self.dispatch_result(assignment.value)
+        self._resolve_assign_targets(assignment, access)
+        return access
 
-        for assign in assignment.to:
-            if assign is None or assign.targets is None:
+    def _resolve_assign_targets(self, assignment: Assignment, access: Access):
+        for assign in assignment.targets:
+            if assign is None:
                 continue
-            elif isinstance(assign.targets, PendingAccess):
-                # Single target, we're good to proceed normally
-                targets = [assign.targets]
-            else:
-                # Multiple nested targets, only overwrite assigned names
-                access = None
-                targets = assign.targets
 
-            for target in targets:
-                if target is None:
-                    continue
+            # On multiple nested targets, only overwrite assigned names
+            value = access if len(assign.elements) <= 1 else None
 
-                if len(target.components) == 1:
-                    comp = Component.from_ast(target.components[0])
-                    self._overwrite(comp.name)
-                    if access is not None:
-                        self._assign(comp.name, access.full_components)
-                else:
-                    self._access(target)
+            for target in assign.elements:
+                self._resolve_assign_target(target, value)
 
-    def _access_simple(self, name: str, lineno: int) -> Optional[Access]:
-        component = Component(name, lineno, lineno, 'load')
-        prior = self.pseudo_scopes_stack[-1].get(component.name, None)
-
-        if prior is None:
+    def _resolve_assign_target(
+        self, target: Optional[PendingAccess], value: Optional[Access]
+    ):
+        if target is None:
             return
 
-        access = Access(LinkContext.none, prior, [component])
-        self.accessed.append(access)
-        return access
+        if len(target.components) == 1:
+            comp = target.components[0]
+            if value is None:
+                self.overwrite_name(comp.name)
+            else:
+                self.assign_name(comp.name, value.full_components)
+                self.create_access(comp.name, target.components)
+        else:
+            self.resolve_pending_access(target)
+
+    def create_simple_access(self, name: str, lineno: int) -> None:
+        """Create single-component access to scope."""
+        component = Component(name, lineno, lineno, 'load')
+        self.create_access(component.name, [component])
+
+    def dispatch_result(
+        self, result: Union[PendingAccess, Assignment, None]
+    ) -> Optional[Access]:
+        """Determine the appropriate processing after tracking an access chain."""
+        if isinstance(result, Assignment):
+            return self.resolve_assignment(result)
+        elif isinstance(result, PendingAccess):
+            return self.resolve_pending_access(result)
 
     def visit_Global(self, node: ast.Global):
         """Import from top scope."""
@@ -331,20 +387,20 @@ class ImportTrackerVisitor(ast.NodeVisitor):
 
         imports = self.outer_scopes_stack[0]
         for name in node.names:
-            self._overwrite(name)
+            self.overwrite_name(name)
             if name in imports:
-                self._assign(name, imports[name])
-                self._access_simple(name, node.lineno)
+                self.assign_name(name, imports[name])
+                self.create_simple_access(name, node.lineno)
 
     def visit_Nonlocal(self, node: ast.Nonlocal):
         """Import from intermediate scopes."""
         imports_stack = self.outer_scopes_stack[1:]
         for name in node.names:
-            self._overwrite(name)
+            self.overwrite_name(name)
             for imports in imports_stack[::-1]:
                 if name in imports:
-                    self.pseudo_scopes_stack[-1][name] = imports[name]
-                    self._access_simple(name, node.lineno)
+                    self.assign_name(name, imports[name])
+                    self.create_simple_access(name, node.lineno)
                     break
 
     def visit_Import(self, node: Union[ast.Import, ast.ImportFrom], prefix: str = ''):
@@ -370,21 +426,20 @@ class ImportTrackerVisitor(ast.NodeVisitor):
             import_names = [name.name for name in node.names]
             aliases = [name.asname for name in node.names]
 
-        end_lineno = getattr(node, 'end_lineno', node.lineno)
         prefix_parts = prefix.rstrip('.').split('.') if prefix else []
         prefix_components = [
-            Component(n, node.lineno, end_lineno, 'load') for n in prefix_parts
+            Component(n, *linenos(node), 'load') for n in prefix_parts
         ]
         if prefix:
-            self.accessed.append(Access(LinkContext.import_from, [], prefix_components))
+            self.save_access(Access(LinkContext.import_from, [], prefix_components))
 
         for import_name, alias in zip(import_names, aliases):
             if not import_star:
                 components = [
-                    Component(n, node.lineno, end_lineno, 'load')
+                    Component(n, *linenos(node), 'load')
                     for n in import_name.split('.')
                 ]
-                self.accessed.append(
+                self.save_access(
                     Access(LinkContext.import_target, [], components, prefix_components)
                 )
 
@@ -394,30 +449,30 @@ class ImportTrackerVisitor(ast.NodeVisitor):
                 import_name = import_name.split('.')[0]
 
             full_components = [
-                Component(n, node.lineno, end_lineno, 'store')
+                Component(n, *linenos(node), 'store')
                 for n in (prefix + import_name).split('.')
             ]
-            self._assign(alias or import_name, full_components)
+            self.assign_name(alias or import_name, full_components)
 
     def visit_ImportFrom(self, node: ast.ImportFrom):
         """Register import source."""
         if node.level:  # relative import
             for name in node.names:
-                self._overwrite(name.asname or name.name)
+                self.overwrite_name(name.asname or name.name)
         else:
             self.visit_Import(node, prefix=node.module + '.')
 
     @track_parents
-    def visit_Name(self, node):
+    def visit_Name(self, node: ast.Name):
         """Visit a Name node."""
-        return PendingAccess([node])
+        return PendingAccess([Component.from_ast(node)])
 
     @track_parents
-    def visit_Attribute(self, node):
+    def visit_Attribute(self, node: ast.Attribute):
         """Visit an Attribute node."""
         inner: Optional[PendingAccess] = self.visit(node.value)
         if inner is not None:
-            inner.components.append(node)
+            inner.components.append(Component.from_ast(node))
         return inner
 
     @track_parents
@@ -425,7 +480,7 @@ class ImportTrackerVisitor(ast.NodeVisitor):
         """Visit a Call node."""
         inner: Optional[PendingAccess] = self.visit(node.func)
         if inner is not None:
-            inner.components.append(node)
+            inner.components.append(Component.from_ast(node))
         with self.reset_parents():
             for arg in node.args + node.keywords:
                 self.visit(arg)
@@ -456,21 +511,31 @@ class ImportTrackerVisitor(ast.NodeVisitor):
     def visit_Assign(self, node: ast.Assign):
         """Visit an Assign node."""
         value = self.visit(node.value)
-        targets = [PendingAssign(self.visit(n)) for n in node.targets[::-1]]
+        targets = []
+        for n in node.targets[::-1]:
+            target = self.visit(n)
+            if not isinstance(target, list):
+                target = [target]
+            targets.append(AssignTarget(target))
         return Assignment(targets, value)
 
     @track_parents
     def visit_AnnAssign(self, node: ast.AnnAssign):
         """Visit an AnnAssign node."""
-        if node.value is not None:
-            value = self.visit(node.value)
+        value = self.visit(node.value) if node.value is not None else None
+        annot = self.visit(node.annotation)
+        if annot is not None:
+            if value is not None:
+                self.resolve_pending_access(value)
+
+            annot.components.append(
+                Component(NameBreak.call, *linenos(node.annotation), 'load')
+            )
+            value = annot
+
         target = self.visit(node.target)
-
-        with self.reset_parents():
-            self.visit(node.annotation)
-
-        if node.value is not None:
-            return Assignment([PendingAssign(target)], value)
+        if value is not None:
+            return Assignment([AssignTarget([target])], value)
 
     def visit_AugAssign(self, node: ast.AugAssign):
         """Visit an AugAssign node."""
@@ -484,7 +549,41 @@ class ImportTrackerVisitor(ast.NodeVisitor):
         """Visit a NamedExpr node."""
         value = self.visit(node.value)
         target = self.visit(node.target)
-        return Assignment([PendingAssign(target)], value)
+        return Assignment([AssignTarget([target])], value)
+
+    @track_parents
+    def visit_MatchClass(self, node):
+        """Visit a match case class as a series of assignments."""
+        with self.reset_parents():
+            cls = self.visit(node.cls)
+
+        accesses = []
+        for n in node.patterns:
+            access = self.visit(n)
+            if access is not None:
+                accesses.append(access)
+
+        assigns = []
+        for attr, pattern in zip(node.kwd_attrs, node.kwd_patterns):
+            target = self.visit(pattern)
+            attr_comps = [
+                Component(NameBreak.call, *linenos(node), 'load'),
+                Component(attr, *linenos(node), 'load'),
+            ]
+            access = PendingAccess(cls.components + attr_comps)
+            assigns.append(Assignment([AssignTarget([target])], access))
+
+        for access in accesses:
+            self.resolve_pending_access(access)
+
+        with self.no_split():
+            for assign in assigns:
+                self.resolve_assignment(assign)
+
+    @track_parents
+    def visit_MatchAs(self, node):
+        """Track match alias names."""
+        return PendingAccess([Component.from_ast(node)])
 
     def visit_AsyncFor(self, node: ast.AsyncFor):
         """Delegate to sync for."""
@@ -508,7 +607,7 @@ class ImportTrackerVisitor(ast.NodeVisitor):
         for kw in node.keywords:
             self.visit(kw)
 
-        self._overwrite(node.name)
+        self.overwrite_name(node.name)
         self.pseudo_scopes_stack.append(self.pseudo_scopes_stack[0].copy())
         for b in node.body:
             self.visit(b)
@@ -525,34 +624,45 @@ class ImportTrackerVisitor(ast.NodeVisitor):
 
     def visit_FunctionDef(self, node: Union[ast.FunctionDef, ast.AsyncFunctionDef]):
         """Swap node order and separate inner scope."""
-        self._overwrite(node.name)
+        self.overwrite_name(node.name)
         for dec in node.decorator_list:
             self.visit(dec)
-        if node.returns is not None:
-            self.visit(node.returns)
         for d in node.args.defaults + node.args.kw_defaults:
             if d is None:
                 continue
             self.visit(d)
         args = self._get_args(node.args)
         args += [node.args.vararg, node.args.kwarg]
-        for arg in args:
-            if arg is None or arg.annotation is None:
-                continue
-            self.visit(arg.annotation)
 
         inner = self.__class__(self.doctree_node)
         inner.pseudo_scopes_stack[0] = self.pseudo_scopes_stack[0].copy()
         inner.outer_scopes_stack = list(self.outer_scopes_stack)
         inner.outer_scopes_stack.append(self.pseudo_scopes_stack[0])
+
         for arg in args:
             if arg is None:
                 continue
-            inner._overwrite(arg.arg)
+            inner.visit(arg)
+        if node.returns is not None:
+            self.visit(node.returns)
         for n in node.body:
             inner.visit(n)
         self.accessed.extend(inner.accessed)
 
+    @track_parents
+    def visit_arg(self, arg: ast.arg):
+        """Handle function argument and its annotation."""
+        target = PendingAccess([Component.from_ast(arg)])
+        if arg.annotation is not None:
+            value = self.visit(arg.annotation)
+            if value is not None:
+                value.components.append(
+                    Component(NameBreak.call, *linenos(arg), 'load')
+                )
+        else:
+            value = None
+        return Assignment([AssignTarget([target])], value)
+
     def visit_Lambda(self, node: ast.Lambda):
         """Swap node order and separate inner scope."""
         for d in node.args.defaults + node.args.kw_defaults:
@@ -567,7 +677,7 @@ class ImportTrackerVisitor(ast.NodeVisitor):
         for arg in args:
             if arg is None:
                 continue
-            inner._overwrite(arg.arg)
+            inner.overwrite_name(arg.arg)
         inner.visit(node.body)
         self.accessed.extend(inner.accessed)
 
