@@ -1,7 +1,6 @@
-import sys
 import pytest
 
-from ._util import refs_equal
+from ._util import refs_equal, skip_walrus
 
 
 class TestAssign:
@@ -17,10 +16,16 @@ class TestAssign:
         refs = [('a', 'a')]
         return s, refs
 
+    @refs_equal
+    def test_assign_to_other_name_linked(self):
+        s = 'import a\nb = a'
+        refs = [('a', 'a'), ('a', 'a'), ('a', 'b')]
+        return s, refs
+
     @refs_equal
     def test_assign_uses_and_assigns_imported(self):
         s = 'import a\na = a\na'
-        refs = [('a', 'a'), ('a', 'a'), ('a', 'a')]
+        refs = [('a', 'a'), ('a', 'a'), ('a', 'a'), ('a', 'a')]
         return s, refs
 
     @refs_equal
@@ -41,10 +46,40 @@ class TestAssign:
         refs = [('a', 'a'), ('a', 'a')]
         return s, refs
 
+    @refs_equal
+    def test_multitarget_assign_to_different_names(self):
+        s = 'import a\nc = b = a'
+        refs = [('a', 'a'), ('a', 'a'), ('a', 'b'), ('a', 'c')]
+        return s, refs
+
     @refs_equal
     def test_multitarget_assign_uses_and_overwrites(self):
         s = 'import a\na = b = a\na, b'
-        refs = [('a', 'a'), ('a', 'a'), ('a', 'a'), ('a', 'b')]
+        refs = [('a', 'a'), ('a', 'a'), ('a', 'b'), ('a', 'a'), ('a', 'a'), ('a', 'b')]
+        return s, refs
+
+    @refs_equal
+    def test_multitarget_assign_overwrites_twice(self):
+        s = 'import a\na = a = a'
+        refs = [('a', 'a'), ('a', 'a'), ('a', 'a'), ('a', 'a')]
+        return s, refs
+
+    @refs_equal
+    def test_multitarget_assign_overwrites_twice_with_attribute(self):
+        s = 'import a\na = a = a.b'
+        refs = [('a', 'a'), ('a.b', 'a.b'), ('a.b', 'a'), ('a.b', 'a')]
+        return s, refs
+
+    @refs_equal
+    def test_multitarget_assign_with_unpacking_first_assigns_to_other(self):
+        s = 'import a\n(c, d) = b = a'
+        refs = [('a', 'a'), ('a', 'a'), ('a', 'b')]
+        return s, refs
+
+    @refs_equal
+    def test_multitarget_assign_with_unpacking_last_assigns_to_other(self):
+        s = 'import a\nd = (b, c) = a'
+        refs = [('a', 'a'), ('a', 'a'), ('a', 'd')]
         return s, refs
 
     @refs_equal
@@ -78,56 +113,86 @@ class TestAssign:
         return s, refs
 
     @refs_equal
-    def test_annassign_uses_imported(self):
+    def test_annassign_overwrites_imported(self):
         s = 'import a\na: b = 1\na'
         refs = [('a', 'a')]
         return s, refs
 
     @refs_equal
     def test_annassign_uses_and_assigns_imported(self):
-        s = 'import a\na: b = a\na'
-        refs = [('a', 'a'), ('a', 'a'), ('a', 'a')]
+        s = 'import a\nb: 1 = a\nb.c'
+        refs = [('a', 'a'), ('a', 'a'), ('a', 'b'), ('a.c', 'b.c')]
+        return s, refs
+
+    @refs_equal
+    def test_annassign_uses_and_annotates_imported(self):
+        s = 'import a\nb: a = 1\nb.c'
+        refs = [('a', 'a'), ('a', 'a'), ('a.()', 'b'), ('a.().c', 'b.c')]
+        return s, refs
+
+    @refs_equal
+    def test_annassign_prioritises_annotation(self):
+        s = 'import a, b\nc: a = b\nc.d'
+        # note that AnnAssign is executed from value -> annot -> target
+        refs = [
+            ('a', 'a'),
+            ('b', 'b'),
+            ('b', 'b'),
+            ('a', 'a'),
+            ('a.()', 'c'),
+            ('a.().d', 'c.d'),
+        ]
         return s, refs
 
     @refs_equal
     def test_annassign_why_would_anyone_do_this(self):
-        s = 'import a\na: a = a\na'
-        refs = [('a', 'a'), ('a', 'a'), ('a', 'a'), ('a', 'a')]
+        s = 'import a\na: a = a\na.b'
+        refs = [('a', 'a'), ('a', 'a'), ('a', 'a'), ('a.()', 'a'), ('a.().b', 'a.b')]
         return s, refs
 
     @refs_equal
     def test_annassign_without_value_overrides_annotation_but_not_linked(self):
+        # note that this is different from runtime behavior
+        # which does not overwrite the variable value
         s = 'import a\na: b\na'
-        refs = [('a', 'a'), ('a', 'a')]
+        refs = [('a', 'a')]
         return s, refs
 
-    @pytest.mark.skipif(
-        sys.version_info < (3, 8), reason='Walrus introduced in Python 3.8.'
-    )
+    @skip_walrus
     @refs_equal
     def test_walrus_uses_imported(self):
         s = 'import a\n(a := 1)\na'
         refs = [('a', 'a')]
         return s, refs
 
-    @pytest.mark.skipif(
-        sys.version_info < (3, 8), reason='Walrus introduced in Python 3.8.'
-    )
+    @skip_walrus
     @refs_equal
     def test_walrus_uses_and_assigns_imported(self):
         s = 'import a\n(a := a)\na'
-        refs = [('a', 'a'), ('a', 'a'), ('a', 'a')]
+        refs = [('a', 'a'), ('a', 'a'), ('a', 'a'), ('a', 'a')]
         return s, refs
 
-    @pytest.mark.skipif(
-        sys.version_info < (3, 8), reason='Walrus introduced in Python 3.8.'
-    )
+    @skip_walrus
     @refs_equal
     def test_walrus_uses_and_assigns_modified_imported(self):
         s = 'import a\n(a := a + 1)\na'
         refs = [('a', 'a'), ('a', 'a')]
         return s, refs
 
+    @skip_walrus
+    @refs_equal
+    def test_nested_walrus_statements(self):
+        s = 'import a\n(c := (b := a))'
+        refs = [('a', 'a'), ('a', 'a'), ('a', 'b'), ('a', 'c')]
+        return s, refs
+
+    @skip_walrus
+    @refs_equal
+    def test_walrus_result_assigned(self):
+        s = 'import a\nc = (b := a)'
+        refs = [('a', 'a'), ('a', 'a'), ('a', 'b'), ('a', 'c')]
+        return s, refs
+
     @refs_equal
     def test_dotted_import_overwrites_all_components(self):
         s = 'class a:\n  b = 1\nimport a.b\na.b'
@@ -146,48 +211,50 @@ class TestFollowAssignment:
     @refs_equal
     def test_follow_simple_assign(self):
         s = 'import a\nb = a\nb'
-        refs = [('a', 'a'), ('a', 'a'), ('a', 'b')]
+        refs = [('a', 'a'), ('a', 'a'), ('a', 'b'), ('a', 'b')]
         return s, refs
 
     @refs_equal
     def test_follow_simple_assign_attr(self):
         s = 'import a\nb = a\nb.c'
-        refs = [('a', 'a'), ('a', 'a'), ('a.c', 'b.c')]
+        refs = [('a', 'a'), ('a', 'a'), ('a', 'b'), ('a.c', 'b.c')]
         return s, refs
 
     @refs_equal
     def test_follow_attr_assign(self):
         s = 'import a\nc = a.b\nc'
-        refs = [('a', 'a'), ('a.b', 'a.b'), ('a.b', 'c')]
+        refs = [('a', 'a'), ('a.b', 'a.b'), ('a.b', 'c'), ('a.b', 'c')]
         return s, refs
 
     @refs_equal
     def test_follow_attr_assign_attr(self):
         s = 'import a\nc = a.b\nc.d'
-        refs = [('a', 'a'), ('a.b', 'a.b'), ('a.b.d', 'c.d')]
+        refs = [('a', 'a'), ('a.b', 'a.b'), ('a.b', 'c'), ('a.b.d', 'c.d')]
         return s, refs
 
     @refs_equal
     def test_follow_attr_call_assign_attr(self):
         s = 'import a\nc = a.b()\nc.d'
-        refs = [('a', 'a'), ('a.b', 'a.b'), ('a.b.().d', 'c.d')]
+        refs = [('a', 'a'), ('a.b', 'a.b'), ('a.b.()', 'c'), ('a.b.().d', 'c.d')]
         return s, refs
 
     @refs_equal
     def test_follow_attr_call_assign_attr_call(self):
         s = 'import a\nc = a.b()\nc.d()'
-        refs = [('a', 'a'), ('a.b', 'a.b'), ('a.b.().d', 'c.d')]
+        refs = [('a', 'a'), ('a.b', 'a.b'), ('a.b.()', 'c'), ('a.b.().d', 'c.d')]
         return s, refs
 
     @refs_equal
     def test_follow_through_two_complex_assignments(self):
-        s = 'import a\nd = a.b().c\ne = d().f'
+        s = 'import a\nd = a.b().c\nf = d().e'
         refs = [
             ('a', 'a'),
             ('a.b', 'a.b'),
             ('a.b.().c', 'c'),
             ('a.b.().c', 'd'),
-            ('a.b.().c.().f', 'f'),
+            ('a.b.().c', 'd'),
+            ('a.b.().c.().e', 'e'),
+            ('a.b.().c.().e', 'f'),
         ]
         return s, refs
 
