@@ -1,7 +1,6 @@
-import sys
 import pytest
 
-from ._util import refs_equal
+from ._util import refs_equal, skip_type_union, skip_walrus
 
 
 class TestFunction:
@@ -42,14 +41,40 @@ class TestFunction:
         return s, refs
 
     @refs_equal
-    def test_func_annotations_then_assigns(self):
-        s = 'import a\ndef f(arg: a) -> a:\n  a = 1'
-        refs = [('a', 'a'), ('a', 'a'), ('a', 'a')]
+    def test_func_annotates_then_uses(self):
+        s = 'import a\ndef f(arg: a):\n  arg.b'
+        refs = [('a', 'a'), ('a', 'a'), ('a.()', 'arg'), ('a.().b', 'arg.b')]
         return s, refs
 
     @refs_equal
-    def test_func_kw_default_uses(self):
-        s = 'import a\ndef f(*_, c, b=a):\n  pass'
+    def test_func_annotates_then_assigns(self):
+        # Note: inner nodes after outer ones
+        s = 'import a\ndef f(arg: a) -> a:\n  a = 1'
+        refs = [('a', 'a'), ('a', 'a'), ('a', 'a'), ('a.()', 'arg')]
+        return s, refs
+
+    @refs_equal
+    def test_func_annotates_as_generic_then_uses(self):
+        s = 'import a\ndef f(arg: a[0]):\n  arg.b'
+        refs = [('a', 'a'), ('a', 'a')]
+        return s, refs
+
+    @refs_equal
+    def test_func_annotates_inside_generic_then_uses(self):
+        s = 'import a\ndef f(arg: b[a]):\n  arg.b'
+        refs = [('a', 'a'), ('a', 'a')]
+        return s, refs
+
+    @skip_type_union
+    @refs_equal
+    def test_func_annotates_union_then_uses(self):
+        s = 'import a\ndef f(arg: a | 1):\n  arg.b'
+        refs = [('a', 'a'), ('a', 'a')]
+        return s, refs
+
+    @refs_equal
+    def test_func_kw_default_uses_not_assigned(self):
+        s = 'import a\ndef f(*_, c, b=a):\n  b'
         refs = [('a', 'a'), ('a', 'a')]
         return s, refs
 
@@ -85,7 +110,7 @@ class TestFunction:
         return s, refs
 
     @refs_equal
-    def test_lambda_arg_default_uses(self):
+    def test_lambda_arg_default_uses_not_assigned(self):
         s = 'import a\nlambda x=a: x'
         refs = [('a', 'a'), ('a', 'a')]
         return s, refs
@@ -233,10 +258,8 @@ class TestComprehension:
         refs = [('a', 'a'), ('a', 'a')]
         return s, refs
 
-    @pytest.mark.skipif(
-        sys.version_info < (3, 8), reason='Walrus introduced in Python 3.8.'
-    )
-    @pytest.mark.xfail(reason='Assignments are not tracked.')
+    @skip_walrus
+    @pytest.mark.xfail(reason='Assignments are not tracked outside of a "scope".')
     @refs_equal
     def test_comp_leaks_walrus(self):
         s = 'import a\n[a := i for i in b]\na'
